import aiohttp
import random
from . import translatorabc, translateresult, constants

# https://github.com/ssut/py-googletrans/issues/268#issue-777566195
DEFAULT_ENDPOINTS = ["https://clients5.google.com/translate_a/t?client=dict-chrome-ex"] 

class GoogleTranslate(translatorabc.TranslatorABC):
    """
    Google Translate API, by default it'll use random API endpoint.
    @endpoint_url: Google Translate API to use.
    @endpoint_urls: Google Translate API list (not recommended), if `endpoint_url` is not specified then it'll use random API endpoint as default.
    """
    def __init__(self, endpoint_url: str = None, endpoint_list: list() = None):
        if endpoint_list is None:
            self._get_default_endpoints()
            endpoint_list = DEFAULT_ENDPOINTS
        self.endpoint_list = endpoint_list
        if endpoint_url is None:
            endpoint_url = endpoint_list[random.randint(0, len(DEFAULT_ENDPOINTS) - 1)]
        self.endpoint_url = endpoint_url

    def _get_default_endpoints(self):
        """
        Get default endpoints defined in the module.
        return set: The default endpoints.
        """
        for endpoint in constants.GOOGLE_TRANSLATE_ENDPOINTS:
            endpoint_url = f"https://{endpoint}/translate_a/single?client=gtx&dt=t"
            if endpoint_url not in DEFAULT_ENDPOINTS:
                DEFAULT_ENDPOINTS.append(endpoint_url)
        return DEFAULT_ENDPOINTS # For convenience, I won't use it btw.

    def _parse_result(self, result: dict):
        """
        Default parser for translated result return from the server, and return a TranslateResult object.
        @result: The JSON-decoded response from the server.
        """
        translate_result = {}
        try:
            translate_result |= {
                "translated_text": result[0][0][0],
                "original_text": result[0][0][1],
                "original_lang": result[2],
                "raw": result
            }
        except (KeyError, TypeError):
            # Use legacy Chrome Translate extension decoder
            sentence = result["sentences"][0]
            translate_result |= {
                "translated_text": sentence["trans"],
                "original_text": sentence["orig"],
                "original_lang": sentence["orig"],
                "raw": result
            }
        translate_result |= {
            "endpoint": self.endpoint_url,
        }
        return translateresult.TranslateResult(**translate_result)

    async def translate(self, text: str, target_language: str, source_language: str = "auto"):
        """
        Translate text from source language to target language
        @text: The text to translate.
        @target_language: The target language to translate the text into.
        @source_language: The source language of the text.
        """
        async with aiohttp.ClientSession() as session:
            async with session.get(self.endpoint_url, params = {
                "q": text,
                "sl": source_language,
                "tl": target_language
            }, headers=constants.GOOGLE_USER_AGENT) as response:
                if response.status == 200:
                    return self._parse_result(await response.json())
                else:
                    raise RuntimeError("Google Translate returned status code {}".format(response.status))

    async def detect(self, text: str):
        """
        Detect the language of the text.
        @text: The text to detect the language.
        """
        # Think smart not work hard :))
        result = await self.translate(text, "en")
        return result.original_lang



LIBRETRANS_HEADER = {
    "Accept": "application/json",
    "Content-Type": "application/x-www-form-urlencoded"
}

# Magic Google Chrome UA
GOOGLE_USER_AGENT = {"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/93.0.4577.82 Safari/537.36"}

# Google Translate servers
# Copied from https://github.com/ssut/py-googletrans/pull/255/files
GOOGLE_TRANSLATE_ENDPOINTS = ['translate.google.ac', 'translate.google.ad', 'translate.google.ae',
                        'translate.google.al', 'translate.google.am', 'translate.google.as',
                        'translate.google.at', 'translate.google.az', 'translate.google.ba',
                        'translate.google.com.ai', 'translate.google.com.ar', 'translate.google.com.au',
                        'translate.google.com.bd', 'translate.google.com.bh', 'translate.google.com.bn',
                        'translate.google.com.bo', 'translate.google.com.br', 'translate.google.com.bz',
                        'translate.google.com.co', 'translate.google.com.cu', 'translate.google.com.cy',
                        'translate.google.com.do', 'translate.google.com.ec', 'translate.google.com.eg',
                        'translate.google.com.et', 'translate.google.com.fj', 'translate.google.com.gh',
                        'translate.google.com.gi', 'translate.google.com.gt', 'translate.google.com.hk',
                        'translate.google.com.jm', 'translate.google.com.kh', 'translate.google.com.kw',
                        'translate.google.com.lb', 'translate.google.com.ly', 'translate.google.com.mm',
                        'translate.google.com.mt', 'translate.google.com.mx', 'translate.google.com.my',
                        'translate.google.com.na', 'translate.google.com.ng', 'translate.google.com.ni',
                        'translate.google.com.np', 'translate.google.com.om', 'translate.google.com.pa',
                        'translate.google.com.pe', 'translate.google.com.pg', 'translate.google.com.ph',
                        'translate.google.com.pk', 'translate.google.com.pr', 'translate.google.com.py',
                        'translate.google.com.qa', 'translate.google.com.sa', 'translate.google.com.sb',
                        'translate.google.com.sg', 'translate.google.com.sl', 'translate.google.com.sv',
                        'translate.google.com.tj', 'translate.google.com.tr', 'translate.google.com.tw',
                        'translate.google.com.ua', 'translate.google.com.uy', 'translate.google.com.vc',
                        'translate.google.com.vn', 'translate.google.com', 'translate.google.cv',
                        'translate.google.cz', 'translate.google.de', 'translate.google.dj',
                        'translate.google.dk', 'translate.google.dm', 'translate.google.dz',
                        'translate.google.st', 'translate.google.td', 'translate.google.tg',
                        'translate.google.tk', 'translate.google.tl', 'translate.google.tm',
                        'translate.google.tn', 'translate.google.to', 'translate.google.tt',
                        'translate.google.us', 'translate.google.vg', 'translate.google.vu',
                        'translate.google.ws']


from abc import ABC, abstractmethod 

class TranslatorABC(ABC):
    """
    Abstract class for translators, you probably don't want to use this unless you're planning to implement a translator API.
    """
    @abstractmethod
    def __init__(self, endpoint_url: str = None, endpoint_urls: list() = None):
        pass

    @abstractmethod
    def _parse_result(self, result: dict):
        pass

    @abstractmethod
    async def translate(self, text: str, target_language: str, source_language: str = "auto"):
        """
        Translate text.
        """
        pass

    @abstractmethod
    async def detect(self, text: str):
        """
        Detect language of text.
        """
        pass
    


import aiohttp
from . import translatorabc, translateresult, constants

# libretranslate.com hasn't implemented rate limiting yet, so you must use an alternate instance or have an API key.

class LibreTranslate(translatorabc.TranslatorABC):
    """
    LibreTranslate API, by default it'll use https://libretranslate.com/
    @endpoint_url: LibreTranslate API to use.
    """
    def __init__(self, endpoint_url: str = "https://translate.argosopentech.com/"):
        self.endpoint_url = endpoint_url

    def _parse_result(self, result: dict):
        """
        Default parser for translated result return from the server, and return a TranslateResult object.
        @result: The JSON-decoded response from the server.
        """
        translate_result = {
            "translated_text": result["translatedText"],
            "original_text": result["original_text"],
            "original_lang": result["original_lang"],
            "endpoint": self.endpoint_url,
            "raw": result
        }
        return translateresult.TranslateResult(**translate_result)

    async def translate(self, text: str, target_language: str, source_language: str = "auto", format_text: str = "text"):
        """
        Translate text from source language to target language
        @text: The text to translate.
        @target_language: The target language to translate the text into.
        @source_language: The source language of the text.
        """
        if source_language == "auto":
            source_language = await self.detect(text)

        async with aiohttp.ClientSession() as session:
            async with session.post(f"{self.endpoint_url}/translate", headers = constants.LIBRETRANS_HEADER, data = {
                "q": text,
                "source": source_language,
                "target": target_language,
                "format": format_text
            }) as response:
                if response.status == 200:
                    rsp = await response.json()
                    rsp |= {
                        "original_text": text,
                        "original_lang": source_language
                    }
                    return self._parse_result(rsp)
                else:
                    raise RuntimeError("LibreTranslate returned status code {}".format(response.status))

    async def detect(self, text: str):
        """
        Detect the language of the text.
        @text: The text to detect the language.
        """
        async with aiohttp.ClientSession() as session:
            async with session.post(f"{self.endpoint_url}/detect", headers = constants.LIBRETRANS_HEADER, data = {
                "q": text
            }) as response:
                if response.status == 200:
                    return (await response.json())[0]["language"]
                else:
                    raise RuntimeError("LibreTranslate returned status code {}".format(response.status))




class TranslateResult():
    """
    This class contain translation result.
    @translated_text: Translated text.
    @original_text: Original text to translate.
    @original_lang: Translated text language.
    @endpoint: Endpoint used to translate.
    @raw: Raw result returned from endpoint (sometimes deobfuscated).
    @raw_obfuscated: Raw obfuscated result returned from endpoint if @raw is deobfuscated).
    """
    def __init__(self, translated_text, original_text, original_lang, endpoint, raw, 
    raw_obfuscated = None, interjection_original = None, interjection_translated = None, original_lang_score: float = None):
        self.translated_text = translated_text
        self.original_text = original_text
        self.original_lang = original_lang
        self.original_lang_score = original_lang_score
        self.endpoint = endpoint
        self.interjection_original = interjection_original
        self.interjection_translated = interjection_translated
        self.raw = raw
        self.raw_obfuscated = raw_obfuscated



"""
Another translate API for Python.
"""
from . import googletranslate, googlewebtranslate, libretranslate, bingtranslator, translateresult

__version__ = "0.2.1"
__author__ = "tretrauit"

# So we don't need to import each translation module individually
GoogleTranslate = googletranslate.GoogleTranslate
GoogleWebTranslate = googlewebtranslate.GoogleWebTranslate
LibreTranslate = libretranslate.LibreTranslate
BingTranslator = bingtranslator.BingTranslator
TranslateResult = translateresult.TranslateResult



import aiohttp
import re
import json
import asyncio
from . import translatorabc, translateresult, constants

DEFAULT_ENDPOINT = "https://www.bing.com/ttranslatev3"

BING_HEADERS = constants.GOOGLE_USER_AGENT | {
    "Host": "www.bing.com",
    "Accept": "*/*",
    "Accept-Language": "en-US,en;q=0.5",
    "Accept-Encoding": "gzip, deflate",
    "Referer": "https://www.bing.com/",
    "Content-Type": "application/x-www-form-urlencoded",
    "Connection": "keep-alive"
}

class BingTranslator(translatorabc.TranslatorABC):
    """
    Bing Microsoft Translator, it'll use https://www.bing.com/ttranslatev3 as the endpoint (and https://www.bing.com/translator for the session).
    """
    def __init__(self):
        self.endpoint_url = DEFAULT_ENDPOINT
        asyncio.run(self._init())
        
    # Asynchronous method to fully initialize BingTranslator
    async def _init(self):
        session_info = None
        async with aiohttp.ClientSession() as session:
            async with session.get("https://www.bing.com/translator", headers = constants.GOOGLE_USER_AGENT) as response:
                if response.status == 200:
                    self.cookies = response.cookies
                    session_info = await response.text()
                else:
                    raise RuntimeError("Bing Translator returned status code {}".format(response.status))
        try:
            # Hey Macrohard using ASP.NET for this, I'm not sure if it's the best way to do.
            self.params = {
                "IG": re.search(',IG:"\w(.+?)"', session_info).group(1),
                "IID": re.search('data-iid="\w(.+?)"', session_info).group(1)
            }
            credentials = json.loads(re.search('params_RichTranslateHelper = (.+?);', session_info).group(1))
            self.token = credentials[1]
            self.key = credentials[0]
        except Exception as e:
            raise RuntimeError(f"Failed to get Bing Translator session info,\n{e}")

    def _parse_result(self, result: dict):
        """
        Default parser for translated result return from the server, and return a TranslateResult object.
        @result: The JSON-decoded response from the server.
        """
        translation = result[0]["translations"][0]
        detected_language = result[0]["detectedLanguage"]
        translate_result = {
            "translated_text": translation["text"],
            "original_text": result[1],
            "original_lang": detected_language["language"],
            "original_lang_score": detected_language["score"],
            "endpoint": self.endpoint_url,
            "raw": result
        }
        return translateresult.TranslateResult(**translate_result)

    async def translate(self, text: str, target_language: str, source_language: str = "auto-detect"):
        """
        Translate text from source language to target language
        @text: The text to translate.
        @target_language: The target language to translate the text into.
        @source_language: The source language of the text.
        """

        data = {
            "text": text,
            "fromLang": source_language,
            "to": target_language,
            "token": self.token,
            "key": self.key
        }
        async with aiohttp.ClientSession() as session:
            async with session.post(f"{self.endpoint_url}", headers = BING_HEADERS, data = data, 
            params = self.params, cookies = self.cookies) as response:
                if response.status == 200:
                    rsp = await response.json() + [text]
                    return self._parse_result(rsp)
                else:
                    raise RuntimeError("Bing Translator returned status code {}".format(response.status))

    async def detect(self, text: str):
        """
        Detect the language of the text.
        @text: The text to detect the language.
        """
        # Think smart not work hard :))
        result = await self.translate(text, "en")
        return result.original_lang



import aiohttp
import random
import json
from . import translatorabc, translateresult, constants

class GoogleWebTranslate(translatorabc.TranslatorABC):
    """
    Google Translate API (used in https://translate.google.com), by default it'll use random API endpoint.
    @endpoint_url: Google Translate endpoint to use.
    @endpoint_urls: Google Translate endpoint list (not recommended), if `endpoint_url` is not specified then it'll use random endpoint as default.
    """
    def __init__(self, endpoint_url: str = None, endpoint_list: list() = None):
        if endpoint_list is None:
            endpoint_list = constants.GOOGLE_TRANSLATE_ENDPOINTS
        self.endpoint_list = endpoint_list
        if endpoint_url is None:
            endpoint_url = endpoint_list[random.randint(0, len(self.endpoint_list) - 1)]
        self.endpoint_url = endpoint_url
        self.RPC_ID = "MkEWBc" # Why this ID works...

    # Copied from https://github.com/ssut/py-googletrans/pull/255/files
    def _build_rpc_request(self, text: str, dest: str, src: str):
        return json.dumps([[
            [
                self.RPC_ID,
                json.dumps([[text, src, dest, True],[None]], separators=(',', ':')),
                None,
                'generic',
            ],
        ]], separators=(',', ':'))

    # Get random endpoints
    def _get_endpoint(self):
        self.endpoint_list[random.randint(0, len(self.endpoint_list) - 1)]

    # Copied from https://github.com/ssut/py-googletrans/pull/255/files with some minimal optimization
    def _parse_result(self, result: str, source_language: str = "auto"):
        """
        Default parser for translated result return from the server, and return a TranslateResult object.
        @result: The fucked JSON response from the server (Good JSON obfuscator there Google ;))
        """
        token_found = False
        square_bracket_counts = [0, 0]
        resp = ''
        for line in result.split("\n"):
            token_found = token_found or f'"{self.RPC_ID}"' in line[:30]

            if not token_found:
                continue

            is_in_string = False
            for i, v in enumerate(line):
                if v == '\"' and line[max(0, i - 1)] != '\\':
                    is_in_string = not is_in_string
                if not is_in_string:
                    if v == '[':
                        square_bracket_counts[0] += 1
                    elif v == ']':
                        square_bracket_counts[1] += 1

            resp += line
            if square_bracket_counts[0] == square_bracket_counts[1]:
                break

        parsed = json.loads(json.loads(resp)[0][2])

        # After decoding using py-googletrans decoder, we get the following JSON like
        # [[None, None, 'vi', [[[0, [[[None, 8]], [True]]]], 8]], [[[None, None, None, True, None, [['Hi', None, None, None, [['Hi', [2]], ['Hello', [2, 5]]]]]]], 'en', 1, 'vi', ['Xin chào', 'auto', 'en', True]], 'vi', ['Xin chào!', None, None, None, None, [[['interjection', [['Hello!', None, ['Alô!', 'Chào anh!', 'Chào chị!', 'Ô nài!', 'Xin chào!'], 1, True]], 'en', 'vi']], 1], None, None, 'vi', 1]]
        # (use https://jsonformatter.curiousconcept.com/ to format it)
        # Then we can try to manually parse it like me below .-.

        #interjection_original = set()
        #interjection_translated = None
        #for v in parsed[3][6][0][0][1][0]:
        #    if isinstance(v, str):
        #        interjection_original.add(v)
        #    elif isinstance(v, list):
        #        interjection_translated = v
        #        break
        original_lang = parsed[0][2]
        if original_lang is None:
            original_lang = source_language

        translate_result = {
            "translated_text": parsed[1][0][0][5][0][0], # What the hell is this
            "original_text": parsed[1][4][0],
            #"interjection_original": list(interjection_original),
            #"interjection_translated": interjection_translated,
            "original_lang": original_lang,
            "endpoint": self.endpoint_url,
            "raw": parsed, # We don't want to give people a mess
            "raw_obfuscated": result # Enjoy hell :(
        }
        return translateresult.TranslateResult(**translate_result)

    # Copied from https://github.com/ssut/py-googletrans/pull/255/files
    async def translate(self, text: str, target_language: str, source_language: str = "auto"):
        """
        Translate text from source language to target language
        @text: The text to translate.
        @target_language: The target language to translate the text into.
        @source_language: The source language of the text.
        """
        data = {
            'f.req': self._build_rpc_request(text, target_language, source_language),
        }
        params = {
            'rpcids': self.RPC_ID,
            'bl': 'boq_translate-webserver_20201207.13_p0',
            'soc-app': 1,
            'soc-platform': 1,
            'soc-device': 1,
            'rt': 'c',
        }
        # Patched headers
        headers = constants.GOOGLE_USER_AGENT | {'Referer': 'https://translate.google.com'}

        async with aiohttp.ClientSession() as session:
            async with session.post("https://{host}/_/TranslateWebserverUi/data/batchexecute".format(host=self.endpoint_url), 
            params=params, data=data, headers=headers) as response:
                if response.status == 200:
                    return self._parse_result(await response.text())
                else:
                    raise RuntimeError("Google Web Translate returned status code {}".format(response.status))

    async def detect(self, text: str):
        """
        Detect the language of the text.
        @text: The text to detect the language.
        """
        # Think smart not work hard :))
        result = await self.translate(text, "en")
        return result.original_lang
